
/*******************************************************************************
*
* File tm1.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Construction of a trivializing stochastic field to a given order in
* perturbation theory.
*
* Syntax: tm1 -i <input file>
*
* For usage instructions see the file README.tm1.
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "utils.h"
#include "contract.h"
#include "plot.h"
#include "version.h"

static int order,iroot;
static int plot,gfix,ghost,ncol;
static theory_t th;
static char theory[NAME_SIZE],nbase[NAME_SIZE];
static char log_dir[NAME_SIZE],tex_dir[NAME_SIZE];
static char log_file[NAME_SIZE],tex_file[NAME_SIZE];
static FILE *fin=NULL,*flog=NULL,*ftex=NULL;


static void set_th(void)
{
   if (strcmp("PHI4",theory)==0)
      th=PHI4;
   else if (strcmp("YM",theory)==0)
      th=YM;
   else if (strcmp("QCD",theory)==0)
      th=QCD;
   else if (strcmp("LYM",theory)==0)
      th=LYM;
   else if (strcmp("LQCD",theory)==0)
      th=LQCD;
   else
      error(1,1,"set_th [tm1.c]","Unknown theory");
}


static void read_theory_parms(void)
{
   find_section(fin,"Theory");
   read_line(fin,"theory","%s",theory);
   read_line(fin,"order","%d",&order);
   read_line(fin,"iroot","%d",&iroot);
   set_th();

   error(order<=0,1,"read_theory_parms [tm1.c]",
         "Parameter order must be positive");

   error((iroot!=0)&&(iroot!=1),1,"read_theory_parms [tm1.c]",
         "Parameter iroot must be 0 or 1");
}


static void read_plot_parms(void)
{
   find_section(fin,"Plotting");
   read_line(fin,"plot","%d",&plot);

   if ((plot==1)&&(th!=PHI4))
   {
      read_line(fin,"gfix","%d",&gfix);
      read_line(fin,"ghost","%d",&ghost);
   }
   else
   {
      gfix=1;
      ghost=1;
   }

   if (plot==1)
   {
      read_line(fin,"ncol","%d",&ncol);

      error(ncol<1,1,"read_plot_parms [tm1.c]",
            "Number of columns must be at least 1");
   }
   else
      ncol=1;
}


static void read_dirs(void)
{
   find_section(fin,"Run name");
   read_line(fin,"name","%s",nbase);

   find_section(fin,"Directories");
   read_line(fin,"log_dir","%s",log_dir);

   if (plot==1)
      read_line(fin,"tex_dir","%s",tex_dir);
   else
      tex_dir[0]='\0';
}


static void setup_files(void)
{
   FILE *ftest;

   check_dir(log_dir);
   error(name_size("%s/%s.tm1.log",log_dir,nbase)>=NAME_SIZE,1,
         "setup_files [tm1.c]","log_dir/<run_name> is too long");
   sprintf(log_file,"%s/%s.tm1.log",log_dir,nbase);
   ftest=fopen(log_file,"r");
   error(ftest!=NULL,1,"setup_files [tm1.c]",
         "Attempt to overwrite old *.log file");

   if (plot==1)
   {
      check_dir(tex_dir);
      error(name_size("%s/%s.tm1.tex",tex_dir,nbase)>=NAME_SIZE,1,
            "setup_files [tm1.c]","tex_dir/<run name> is too long");
      sprintf(tex_file,"%s/%s.tm1.tex",tex_dir,nbase);
      ftest=fopen(tex_file,"r");
      error(ftest!=NULL,1,"setup_files [tm1.c]",
            "Attempt to overwrite old *.tex file");
   }
}


static void read_infile(int argc,char *argv[])
{
   int ifile;

   ifile=find_opt(argc,argv,"-i");
   error((ifile==0)||(ifile==(argc-1)),1,"read_infile [tm1.c]",
         "Syntax: tm1 -i <input file>");
   fin=freopen(argv[ifile+1],"r",stdin);
   error(fin==NULL,1,"read_infile [tm1.c]",
         "Unable to open input file");

   read_theory_parms();
   read_plot_parms();
   read_dirs();
   setup_files();
   fclose(fin);
}


static void print_info(void)
{
   flog=freopen(log_file,"w",stdout);
   error(flog==NULL,1,"print_info [tm1.c]","Unable to open log file");
   printf("\n");

   printf("Construction of a trivializing stochastic field\n");
   printf("-----------------------------------------------\n\n");

   printf("Program version %s\n\n",ISPT_RELEASE);
   printf("Theory = %s\n",theory);
   printf("Order = %d\n",order);
   printf("iroot = %d\n\n",iroot);

   fflush(flog);
}


static void print_dgmno(void)
{
   int i,j,k,l,nlp;
   int sm0[6],sm1[6];
   cnum_t **cna;
   vlist_t vl;
   ltree_t *ltd,**lta;
   tlist_t *tl;

   printf("The numbers of diagrams at fixed order in the coupling are:\n\n");

   printf(" order    all   c(R)!=0    ");

   if (th!=PHI4)
      printf(" w/o gfix  c(R)!=0    w/o gfix&ghost  c(R)!=0\n");
   else
      printf("\n");

   printf("-------------------------");

   if (th!=PHI4)
      printf("-------------------------------------------------\n");
   else
      printf("\n");

   for (l=0;l<6;l++)
      sm0[l]=0;

   vl=vertices(th,order);

   for (i=1;i<=order;i++)
   {
      for (l=0;l<6;l++)
         sm1[l]=0;

      nlp=nlp_max(th,i);

      for (j=0;j<=nlp;j++)
      {
         tl=select_ltrees(th,i,j);

         if (tl!=NULL)
         {
            lta=(*tl).lta;
            cna=(*tl).cna;

            for (k=0;k<(*tl).n;k++)
            {
               ltd=lta[k];

               sm1[0]+=1;
               sm1[1]+=is_zero_cnum(cna[k]);

               if (has_gfix((*ltd).dm,&vl)==0)
               {
                  sm1[2]+=1;
                  sm1[3]+=is_zero_cnum(cna[k]);

                  if (has_ghost((*ltd).dm,&vl)==0)
                  {
                     sm1[4]+=1;
                     sm1[5]+=is_zero_cnum(cna[k]);
                  }
               }
            }
         }
      }

      printf("  %2d  %7d  %7d",i,sm1[0],sm1[0]-sm1[1]);

      if (th!=PHI4)
         printf("     %7d   %7d       %7d      %7d\n",
                sm1[2],sm1[2]-sm1[3],sm1[4],sm1[4]-sm1[5]);
      else
         printf("\n");

      for (l=0;l<6;l++)
         sm0[l]+=sm1[l];
   }

   printf("-------------------------");

   if (th!=PHI4)
      printf("-------------------------------------------------\n");
   else
      printf("\n");

   printf("      %7d  %7d",sm0[0],sm0[0]-sm0[1]);

   if (th!=PHI4)
      printf("     %7d   %7d       %7d      %7d\n\n",
             sm0[2],sm0[2]-sm0[3],sm0[4],sm0[4]-sm0[5]);
   else
      printf("\n\n");
}


static void print_plot(void)
{
   int i,j,k,nlp;
   double pos1[2],pos2[2];
   cnum_t **cna;
   vlist_t vl;
   ltree_t *ltd,**lta;
   tlist_t *tl;

   ftex=fopen(tex_file,"w");
   error(ftex==NULL,1,"print_plot [tm1.c]","Unable to open plot file");

   start_plot(ftex,ncol);
   vl=vertices(th,order);

   for (i=1;i<=order;i++)
   {
      nlp=nlp_max(th,i);

      for (j=0;j<=nlp;j++)
      {
         tl=select_ltrees(th,i,j);

         if (tl!=NULL)
         {
            lta=(*tl).lta;
            cna=(*tl).cna;

            for (k=0;k<(*tl).n;k++)
            {
               ltd=lta[k];

               if (((gfix==1)||(has_gfix((*ltd).dm,&vl)==0))&&
                   ((ghost==1)||(has_ghost((*ltd).dm,&vl)==0)))
               {
                  start_bbox();
                  draw_ltree1(cna[k],lta[k],&vl);
                  end_bbox(pos1,pos2);

                  start_diagram(pos1,pos2);
                  draw_ltree1(cna[k],lta[k],&vl);
                  end_diagram();
               }
            }
         }
      }
   }

   end_plot();
   fclose(ftex);
}


int main(int argc,char *argv[])
{
   int ndm;
   double dt;
   clock_t t1,t2;

   read_infile(argc,argv);
   print_info();

   t1=clock();
   ndm=tmap(th,order,iroot);
   t2=clock();
   dt=(double)(t2-t1)/(double)(CLOCKS_PER_SEC);

   if ((t1!=(clock_t)(-1))&&(t2!=(clock_t)(-1)))
   {
      printf("Trivializing map obtained in %.1e seconds",dt);

      if (dt==0.0)
         printf(" (= less than the clock resolution).\n");
      else
         printf(".\n");
   }

   printf("The total number of tree diagrams generated is %d.\n\n",ndm);
   print_dgmno();

   if (plot==1)
   {
      printf("Now produce LaTeX plot file.\n");

      if (gfix!=1)
      {
         printf("Diagrams with gauge-fixing counterterms ");

         if (ghost!=1)
            printf("and ghost vertices ");

         printf("are omitted.\n");
      }
      else if (ghost!=1)
         printf("Diagrams with ghost vertices are omitted.\n");

      if (ncol>1)
         printf("Try to print diagrams in %d columns.\n",ncol);
      printf("Diagrams with vanishing coefficient are printed too.\n");

      print_plot();
      printf("Plot commands printed to file %s\n",tex_file);
   }

   printf("All done.\n\n");
   fclose(flog);
   exit(0);
}
