
/*******************************************************************************
*
* File contract.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Contraction of products of labeled tree diagrams.
*
* The externally accessible functions are
*
*   long init_contract(int nl,src_legs *sl)
*     Initializes the contraction generator for contractions of 2*nl
*     label=0 and all label>0 sources recorded in the structure sl
*     (see the notes). An error occurs if there are label=0 fermion
*     sources or more than one type of label=0 sources in the source
*     inventory sl. The program returns the number of different line
*     lists generated by next_contract().
*
*   long next_contract(line_t *lnl,long *nc)
*     Assigns the next line list of the type specified by init_contract()
*     to the array lnl and returns the current list count (see the notes).
*     The number of Wick contractions of the sources leading to the line
*     list lnl is assigned to nc. The generator wraps around when all
*     possible line lists have been generated.
*
*   void stop_contract(void)
*     Resets the contraction generator to its uninitialized state. Note
*     that a running generator must be stopped by calling this program
*     before it can be re-initialized.
*
*   void merge_ltrees(int nd,ltree_t **ltd,int nl,line_t *lnl,diagram_t *dm)
*     Forms the diagram dm by adding the list of lines lnl to the union
*     of the tree diagrams ltd[0].dm,..,ltd[nd-1].dm. The vertex indices
*     in the line list are interpreted as the position of the vertices
*     (counting from 0) in the sequence of all vertices of the diagrams.
*
*   long mfact_ltrees(int m,int *n,int *idx)
*     Returns the multiplicity factor k1!*..*kr! of the values in the
*     structured index array idx[0],idx[1],.. (see the notes).
*
* Notes:
*
* The programs in this module serve to contract the source fields attached to
* the vertices of a set ltd[0],..,ltd[nd-1] of labeled tree diagrams. First
* the source inventory sl must be obtained by calling the program src_legs()
* [src_legs.c]. The contraction generator is then initialized and the line
* lists generated are passed to merge_ltrees(), which produces the contracted
* diagrams.
*
* There can be many Wick contractions of the source fields that lead to the
* same line list and therefore to the same contracted diagram. The program
* next_contract() cycles through the set of all line lists rather than all
* Wick contractions.
*
* The number of lines in the line lists lnl generated by next_contract() is
* equal to
*
*  nl+{sum_{i=1}^nt ns[i]}/2,
*
* where
*
*  nt=(*sl).nt,  ns=(*sl).ns,
*
* and nl and sl are the arguments passed to init_contract().
*
* The contraction generator relies on the line list and Wick generators (see
* permutations/lnlist.c and permutations/wickgen.c). These can therefore not
* simultaneously be used for other purposes. Any attempt to do so results in
* an error.
*
* The index array analyzed by the program mfact_ltrees() is logically divided
* into m groups of n[0],..,n[m-1] elements. Within each group the indices
* are ordered in descending order, i.e.
*
*  idx[0]>=idx[1]>=..>=idx[n[0]-1],  idx[n[0]]>=idx[n[0]+1]>=..
*
* The program further divides the indices in each group into ranges of equal
* values and determines the lengths k1,..,kr of these subsequences. Isolated
* indices ..>idx[l]>.. are counted as a sequence of length 1, for example,
* two indices ..>idx[l]=idx[l+1]>.. as a sequence of length 2, etc., where
* each sequence must be fully contained within one group.
*
*******************************************************************************/

#define CONTRACT_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "permutations.h"
#include "contract.h"

static int ma=0,nva=0,nla=0,init=0;
static int mb,mf,nvb,nvf,nlb,nlf;
static int *nvs,*nls,**ivs,**nes;
static int (*lns)[2];
static long np1,np2,npm1,npm2,ncs;
static ltype_t *lts;


static void alloc_arrays(void)
{
   if (((2*mb+mf)>ma)||((2*nvb+nvf)>nva)||((nlb+nlf)>nla))
   {
      if ((ma)>0)
      {
         free(ivs);
         free(nvs);
      }

      if (nla>0)
      {
         free(lns);
         free(lts);
      }

      if ((2*mb+mf)>ma)
         ma=2*mb+mf;
      if ((2*nvb+nvf)>nva)
         nva=2*nvb+nvf;
      if ((nlb+nlf)>nla)
         nla=nlb+nlf;

      ivs=malloc(ma*sizeof(*ivs));
      nvs=malloc((ma+nva)*sizeof(*nvs));
      lns=malloc(nla*sizeof(*lns));
      lts=malloc(nla*sizeof(*lts));

      error((ivs==NULL)||(nvs==NULL)||(lns==NULL)||(lts==NULL),1,
            "alloc_arrays [contract.c]","Unable to allocate auxiliary arrays");
   }
}


static void set_arrays(int nl,src_legs_t *sl)
{
   int nt,*lbl,*nv,*ns,**iv,**ne;
   int i,j,k,kb,kf,lb,lf,ie;
   ltype_t *lt,lr;

   nt=(*sl).nt;
   lbl=(*sl).lbl;
   nv=(*sl).nv;
   ns=(*sl).ns;
   iv=(*sl).iv;
   ne=(*sl).ne;
   lt=(*sl).lt;

   mb=0;
   mf=0;
   nvb=0;
   nvf=0;
   nlb=0;
   nlf=0;
   ie=(nl<0);
   k=0;

   for (i=0;i<nt;i++)
   {
      if ((lt[i]==SCALAR)||(lt[i]==GLUON))
      {
         if (lbl[i]>0)
         {
            ie|=(ns[i]&0x1);

            if (ns[i]>=2)
            {
               nvb+=nv[i];
               nlb+=ns[i]/2;
               mb+=1;
            }
         }
         else
         {
            ie|=(k!=0);
            k=1;

            if (nl>=1)
            {
               nvb+=nv[i];
               nlb+=nl;
               mb+=1;
            }
         }
      }
      else if ((lt[i]==GHOST)||(lt[i]==QUARK))
      {
         if (lbl[i]>0)
         {
            if (ns[i]>=1)
            {
               nvf+=2*nv[i];
               nlf+=ns[i];
               mf+=1;
            }
         }
         else
            ie=1;
      }
   }

   error(ie!=0,1,"set_arrays [contract.c]",
         "Unexpected source inventory");

   if ((mb==0)&&(mf==0))
      return;

   alloc_arrays();

   nes=ivs+mb+mf;
   nls=nvs+mb+mf;
   ivs[0]=nls+mb;
   ivs[mb]=ivs[0]+nvb;
   nes[0]=ivs[mb]+nvf;

   kb=0;
   kf=mb;
   lb=0;
   lf=nlb;

   for (i=0;i<nt;i++)
   {
      if ((lt[i]==SCALAR)||(lt[i]==GLUON))
      {
         if (((lbl[i]>0)&&(ns[i]>=2))||((lbl[i]==0)&&(nl>=1)))
         {
            nvs[kb]=nv[i];

            if (lbl[i]==0)
               nls[kb]=nl;
            else
               nls[kb]=ns[i]/2;

            if (kb>0)
            {
               ivs[kb]=ivs[kb-1]+nvs[kb-1];
               nes[kb]=nes[kb-1]+nvs[kb-1];
            }

            for (j=0;j<nv[i];j++)
            {
               ivs[kb][j]=iv[i][j];
               nes[kb][j]=ne[i][j];
            }

            for (j=0;j<nls[kb];j++)
            {
               lts[lb]=lt[i];
               lb+=1;
            }

            kb+=1;
         }
      }
      else if ((lt[i]==GHOST)||(lt[i]==QUARK))
      {
         if ((lbl[i]>0)&&(ns[i]>=1))
         {
            nvs[kf]=2*nv[i];

            if (kf>mb)
               ivs[kf]=ivs[kf-1]+nvs[kf-1];

            for (j=0;j<nv[i];j++)
               ivs[kf][j]=iv[i][j];

            if (lt[i]==GHOST)
               lr=ANTIGHOST;
            else
               lr=ANTIQUARK;

            for (k=0;k<nt;k++)
            {
               if ((lt[k]==lr)&&(lbl[k]==lbl[i]))
               {
                  for (j=0;j<nv[k];j++)
                     ivs[kf][nv[i]+j]=iv[k][j];

                  break;
               }
            }

            ie|=(k==nt);

            for (j=0;j<ns[i];j++)
            {
               lts[lf]=lt[i];
               lf+=1;
            }

            kf+=1;
         }
      }
   }

   error(ie!=0,2,"set_arrays [contract.c]",
         "Unexpected source inventory");
}


static void assign_lns(line_t *lnl)
{
   int i;

   for (i=0;i<(nlb+nlf);i++)
   {
      lnl[i].type=lts[i];
      lnl[i].iv[0]=lns[i][0];
      lnl[i].iv[1]=lns[i][1];
   }
}


long init_contract(int nl,src_legs_t *sl)
{
   error(init!=0,1,"init_contract [lnlist.c]",
         "A running contraction generator cannot be re-initialized");

   set_arrays(nl,sl);

   if ((mb==0)&&(mf==0))
      return 0L;

   if (mb>0)
   {
      npm1=init_lnlist(mb,nvs,nls,ivs,nes);
      np1=next_lnlist(lns,&ncs);
   }
   else
   {
      npm1=1L;
      np1=0L;
   }

   if (mf>0)
   {
      npm2=init_wick(mf,0,nvs+mb,ivs+mb);
      np2=next_wick(lns+nlb);
   }
   else
   {
      npm2=1L;
      np2=0L;
   }

   error(npm1>(LONG_MAX/npm2),1,"init_contract [contract.c]",
         "Number of contractions is out of range");
   init=1;

   return npm1*npm2;
}


long next_contract(line_t *lnl,long *nc)
{
   long np;

   error(init!=1,1,"next_contract [contract.c]",
         "Contraction generator is not initialized");

   assign_lns(lnl);
   (*nc)=ncs;

   if ((mb>0)&&(mf>0))
   {
      np=npm2*(np1-1L)+np2;
      np2=next_wick(lns+nlb);

      if (np2==1L)
         np1=next_lnlist(lns,&ncs);
   }
   else if (mb>0)
   {
      np=np1;
      np1=next_lnlist(lns,&ncs);
   }
   else if (mf>0)
   {
      np=np2;
      np2=next_wick(lns+nlb);
   }
   else
      np=0L;

   return np;
}


void stop_contract(void)
{
   if (init==1)
   {
      stop_lnlist();
      stop_wick();
      init=0;
   }
}


static void copy_vertices(int n,int (*v1)[2],int (*v2)[2])
{
   int i;

   for (i=0;i<n;i++)
   {
      (*v2)[0]=(*v1)[0];
      (*v2)[1]=(*v1)[1];

      v1+=1;
      v2+=1;
   }
}


static void copy_lines(int n,int s,line_t *l1,line_t *l2)
{
   int i;

   for (i=0;i<n;i++)
   {
      (*l2).type=(*l1).type;
      (*l2).iv[0]=(*l1).iv[0]+s;
      (*l2).iv[1]=(*l1).iv[1]+s;

      l1+=1;
      l2+=1;
   }
}


void merge_ltrees(int nd,ltree_t **ltd,int nl,line_t *lnl,diagram_t *dm)
{
   int i,s,iv0,iv1;
   int snv,snl,(*v)[2];
   line_t *l;
   diagram_t *dml;

   snv=0;
   snl=nl;

   for (i=0;i<nd;i++)
   {
      dml=ltd[i][0].dm;
      snv+=(*dml).nv;
      snl+=(*dml).nl;
   }

   error((snv>(*dm).nv)||(snl>(*dm).nl),1,"merge_ltrees [contract.c]",
         "Arrays in diagram structure are too short");

   for (i=0;i<nl;i++)
   {
      iv0=lnl[i].iv[0];
      iv1=lnl[i].iv[1];

      if ((iv0<0)||(iv0>=snv)||(iv1<0)||(iv1>=snv))
         error(1,1,"merge_ltrees [contract.c]",
               "Vertex indices are out of range");
   }

   (*dm).nv=snv;
   (*dm).nl=snl;
   v=(*dm).v;
   l=(*dm).l;
   s=0;

   for (i=0;i<nd;i++)
   {
      dml=ltd[i][0].dm;

      copy_vertices((*dml).nv,(*dml).v,v);
      copy_lines((*dml).nl,s,(*dml).l,l);

      v+=(*dml).nv;
      l+=(*dml).nl;
      s+=(*dml).nv;
   }

   copy_lines(nl,0,lnl,l);
}


static long acc_fact(int k,long f)
{
   long f1;

   if (k>1)
   {
      f1=factorial(k);

      if (f<=(LONG_MAX/f1))
         f*=f1;
      else
         error(1,1,"acc_fact [contract.c]",
               "Multiplicity factor is out of range");
   }

   return f;
}


long mfact_ltrees(int m,int *n,int *idx)
{
   int i,j,k;
   long f;

   f=1L;

   for (i=0;i<m;i++)
   {
      k=1;

      for (j=1;j<n[i];j++)
      {
         if (idx[j]!=idx[j-1])
         {
            f=acc_fact(k,f);
            k=1;
         }
         else
            k+=1;
      }

      f=acc_fact(k,f);
      idx+=n[i];
   }

   return f;
}
