
/*******************************************************************************
*
* File lnlcon.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Connectivity of line lists.
*
* The externally accessible functions are
*
*   void init_vsets(vsets_t *vs)
*     Initializes the structure vs. Every structure of type vsets_t
*     should be initialized before it is passed as argument to other
*     programs.
*
*   void get_vsets(int nd,ltree_t **ltd,vsets_t *vs)
*     Labels the vertices of the diagrams ltd[0],..,ltd[nd-1] by an index
*     0,1,2,.. and assigns the indices of the vertices of the i'th diagram
*     to the elements of the array (*vs).iv[i]. The numbers of vertices in
*     the diagrams are assigned to (*vs).nv[i] and the number nd of diagrams
*     to (*vs).ns. The size of the arrays in the structure vs is dynamically
*     adjusted when needed.
*
*   int lnlcon(vset_t *vs,int nl,line_t *lnl)
*     Returns 1 if the sets of vertices described by the structure vs are
*     connected by the lines lnl[0],..,lnl[nl-1]. Otherwise 0 is returned.
*
* Notes:
*
* The data in a structure of type vsets_t describe sets of vertices that
* are to be treated as single vertices like the vertices in a skeleton
* diagram. Two sets of vertices are considered to be connected by a line
* if one endpoint of the line is a vertex in the first set and the other
* a vertex in the second set.
*
* The elements of a structure of type vsets_t are:
*
*   nsm      Maximal number of vertex sets that can be accommodated in
*            the structure.
*
*   nvm      Maximal total number of vertices that can be accommodated
*            in the structure.
*
*   ns       Current number of vertex sets in the structure.
*
*   nv       Number nv[i], i=0,..,ns-1, of the vertices in the i'th set.
*
*   iv       Index iv[i][j], j=0,..,nv[i]-1, of the vertex number j in
*            the i'th set.
*
* The vertex sets passed to the program lnlcon() need not contain all the
* endpoints occuring in the line list lnl nor do all vertices in the sets
* have to be endpoints of some lines. The vertices (at least the ones
* occuring in the line list) must however have pairwise different indices.
* An error occurs if this is not so.
*
*******************************************************************************/

#define LNLCON_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "contract.h"

static int nva,nla,imn,imx;
static int *ivs,*ils,(*ps)[2];


void init_vsets(vsets_t *vs)
{
   (*vs).nsm=0;
   (*vs).nvm=0;
   (*vs).ns=0;

   (*vs).nv=NULL;
   (*vs).iv=NULL;
}


static void free_vsets(vsets_t *vs)
{
   if ((*vs).nsm>0)
   {
      free((*vs).nv);
      free((*vs).iv);
   }
}


static void alloc_vsets(int nsm,int nvm,vsets_t *vs)
{
   int *nv,**iv;

   if ((nsm>(*vs).nsm)||(nvm>(*vs).nvm))
   {
      if ((*vs).nsm>nsm)
         nsm=(*vs).nsm;
      if ((*vs).nvm>nvm)
         nvm=(*vs).nvm;

      free_vsets(vs);

      nv=malloc((nsm+nvm)*sizeof(*nv));
      iv=malloc(nsm*sizeof(*iv));

      error((nv==NULL)||(iv==NULL),1,"alloc_vsets [vsets.c]",
            "Unable to allocate data arrays");

      (*vs).nsm=nsm;
      (*vs).nvm=nvm;
      (*vs).nv=nv;
      (*vs).iv=iv;
   }
}


void get_vsets(int nd,ltree_t **ltd,vsets_t *vs)
{
   int nv,nvm;
   int i,j,k;
   diagram_t *dm;

   error(nd<1,1,"get_vsets [lnlcon.c]","Empty list of diagrams");
   nvm=0;

   for (i=0;i<nd;i++)
   {
      dm=ltd[i][0].dm;
      nvm+=(*dm).nv;
   }

   alloc_vsets(nd,nvm,vs);

   (*vs).ns=nd;
   (*vs).iv[0]=(*vs).nv+nd;
   k=0;

   for (i=0;i<nd;i++)
   {
      if (i>0)
         (*vs).iv[i]=(*vs).iv[i-1]+(*vs).nv[i-1];

      dm=ltd[i][0].dm;
      nv=(*dm).nv;
      (*vs).nv[i]=nv;

      for (j=0;j<nv;j++)
      {
         (*vs).iv[i][j]=k;
         k+=1;
      }
   }
}


static void fnd_range(int nl,line_t *lnl)
{
   int i,iv0,iv1;

   iv0=lnl[0].iv[0];
   iv1=lnl[0].iv[1];
   imn=iv0;
   imx=iv0;

   if (iv1<imn)
      imn=iv1;
   if (iv1>imx)
      imx=iv1;

   for (i=1;i<nl;i++)
   {
      iv0=lnl[i].iv[0];
      iv1=lnl[i].iv[1];

      if (iv0<imn)
         imn=iv0;
      if (iv0>imx)
         imx=iv0;

      if (iv1<imn)
         imn=iv1;
      if (iv1>imx)
         imx=iv1;
   }
}


static void alloc_arrays(int ns,int nl,line_t *lnl)
{
   int nv;

   fnd_range(nl,lnl);

   if (ns<(imx-imn+1))
      nv=imx-imn+1;
   else
      nv=ns;

   if ((nv>nva)||(nl>nla))
   {
      if (nva>0)
         free(ivs);
      if (nla>0)
         free(ps);

      if (nv>nva)
         nva=nv;
      if (nl>nla)
         nla=nl;

      ivs=malloc((nva+nla)*sizeof(*ivs));
      ps=malloc(nla*sizeof(*ps));

      error((ivs==NULL)||(ps==NULL),1,"alloc_arrays [lnlcon.c]",
            "Unable to allocate auxiliary arrays");
   }

   ils=ivs+nv;
}


static void set_ivs(int ns,int *nv,int **iv)
{
   int k,l,j;

   for (k=0;k<=(imx-imn);k++)
      ivs[k]=ns;

   for (k=0;k<ns;k++)
   {
      for (l=0;l<nv[k];l++)
      {
         j=iv[k][l];

         if ((imn<=j)&&(j<=imx))
         {
            j-=imn;

            if (ivs[j]==ns)
               ivs[j]=k;
            else
               error(1,1,"set_ivs [lnlcon.c]","Dublicate vertex labels");
         }
      }
   }
}


static int set_ps(int ns,int nl,line_t *lnl)
{
   int n,l,j;
   int i1,i2;

   n=0;

   for (l=0;l<nl;l++)
   {
      i1=ivs[lnl[l].iv[0]-imn];
      i2=ivs[lnl[l].iv[1]-imn];

      if ((i1<ns)&&(i2<ns)&&(i1!=i2))
      {
         if (i1>i2)
         {
            j=i1;
            i1=i2;
            i2=j;
         }

         for (j=0;j<n;j++)
         {
            if ((i1==ps[j][0])&&(i2==ps[j][1]))
               break;
         }

         if (j==n)
         {
            ps[n][0]=i1;
            ps[n][1]=i2;
            n+=1;
         }
      }
   }

   return n;
}


int lnlcon(vsets_t *vs,int nl,line_t *lnl)
{
   int ns,k,l,s;
   int iv1,iv2;
   int *nv,**iv;

   ns=(*vs).ns;
   nv=(*vs).nv;
   iv=(*vs).iv;

   error((ns<1)||(nl<0),1,"lnlcon [lnlcon.c]",
         "Parameters are out of range");

   for (k=0;k<ns;k++)
   {
      if (nv[k]<=0)
         error(1,1,"lnlcon [lnlcon.c]","Empty vertex set");
   }

   if (ns==1)
      return 1;
   if (nl==0)
      return 0;

   alloc_arrays(ns,nl,lnl);
   set_ivs(ns,nv,iv);
   nl=set_ps(ns,nl,lnl);
   ivs[0]=1;

   for (k=1;k<ns;k++)
      ivs[k]=0;

   for (k=0;k<nl;k++)
      ils[k]=0;

   s=1;

   while (s==1)
   {
      s=0;

      for (k=0;k<nl;k++)
      {
         if (ils[k]==0)
         {
            iv1=ps[k][0];
            iv2=ps[k][1];
            l=ivs[iv1]+ivs[iv2];

            if (l==1)
            {
               ivs[iv1]=1;
               ivs[iv2]=1;
               s=1;
            }

            if (l>0)
               ils[k]=1;
         }
      }
   }

   for (k=1;k<ns;k++)
   {
      if (ivs[k]==0)
         return 0;
   }

   return 1;
}
