
/*******************************************************************************
*
* File sfactor.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the symmetry factors of a diagram.
*
* The externally accessible function is
*
*   long sfactor(theory_t th,diagram_t *dm,long *sf)
*     Returns the symmetry factor F1*..*F4 of the diagram dm in the theory
*     with label th and assigns the factors F1,..,F4 to the array elements
*     sf[0],..,sf[3]. An error occurs if the diagram has vertices of type
*     SOURCE or EXTERNAL or if the symmetry factor is larger than LONG_MAX.
*
* Notes:
*
* For diagrams with external lines labeled by the appropriate field indices,
* the symmetry factors F1,..,F4 are defined in the paper
*
*  M. Luescher, "Instantaneous stochastic perturbation theory"
*
* Some further discussion of the symmetry factors is included in the notes
* doc/ISPT-notes.pdf.
*
* The diagram data structures considered here do not include the external
* lines in the list of lines, but the number and type of external lines at
* a given vertex can be inferred from the data (see diagram.c). The program
* sfactor() implicitly assigns distinct field labels to the external lines
* and calculates the symmetry factors defined in the notes.
*
*******************************************************************************/

#define SFACTOR_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "permutations.h"
#include "rvt.h"

static int nvs=0,nls=0;
static int **iln,**eln,*rkv,*ivs;
static int msi,mse,*ns,*ips,*ils;
static line_t *lold,*lnew;


static void alloc_aux(int nv,int nl)
{
   int nt,i;

   if (nvs>0)
   {
      free(iln);
      free(rkv);
      free(lold);
   }

   if (nv<1)
      nv=1;
   if (nl<1)
      nl=1;
   if (nv<nvs)
      nv=nvs;
   if (nl<nls)
      nl=nls;

   nt=(int)(LTYPES);
   iln=malloc(2*nv*sizeof(*iln));
   rkv=malloc((4*nv+nl+2*nv*nt)*sizeof(*rkv));
   lold=malloc(2*nl*sizeof(*lold));

   error((iln==NULL)||(rkv==NULL)||(lold==NULL),1,"alloc_aux [sfactor.c]",
         "Unable to allocate auxiliary arrays");

   ivs=rkv+nv;
   ns=ivs+nv;
   ips=ns+nv;
   ils=ips+nv;

   eln=iln+nv;
   iln[0]=ils+nl;

   for (i=1;i<(2*nv);i++)
      iln[i]=iln[i-1]+nt;

   lnew=lold+nl;
   nvs=nv;
   nls=nl;
}


static void set_xln(theory_t th,diagram_t *dm)
{
   int nv,nl,order,ie;
   int i,j,iv0,iv1;
   int (*v)[2];
   line_t *l;
   vlist_t vls;
   vertex_t *vx;

   nv=(*dm).nv;
   nl=(*dm).nl;
   v=(*dm).v;
   l=(*dm).l;

   if ((nv>nvs)||(nl>nls))
      alloc_aux(nv,nl);

   order=0;
   ie=0;

   for (i=0;i<nv;i++)
   {
      if (v[i][0]>order)
         order=v[i][0];
   }

   vls=vertices(th,order);

   for (i=0;i<nv;i++)
   {
      vx=vls.v[v[i][0]]+v[i][1];
      ie|=(((*vx).type==SOURCE)||((*vx).type==EXTERNAL));

      for (j=0;j<(int)(LTYPES);j++)
      {
         iln[i][j]=0;
         eln[i][j]=(*vx).nl[j];
      }
   }

   error(ie!=0,1,"set_xln [sfactor.c]",
         "Diagram has SOURCE or EXTERNAL vertices");

   for (i=0;i<nl;i++)
   {
      j=(int)(l[i].type);
      iv0=l[i].iv[0];
      iv1=l[i].iv[1];

      iln[iv0][j]+=1;
      iln[iv1][(int)(rvt[j])]+=1;

      eln[iv0][j]-=1;
      eln[iv1][(int)(rvt[j])]-=1;
   }
}


static int cmp_vx(int (*v)[2],int iv0,int iv1)
{
   int j,ne0,ne1;
   int k0,k1,l0,l1;

   for (j=0;j<(int)(LTYPES);j++)
   {
      ne0=eln[iv0][j];
      ne1=eln[iv1][j];

      if (ne0>ne1)
         return 1;
      else if (ne0<ne1)
         return -1;
   }

   k0=v[iv0][0];
   k1=v[iv1][0];

   if (k0>k1)
      return 1;
   else if (k0<k1)
      return -1;

   l0=v[iv0][1];
   l1=v[iv1][1];

   if (l0>l1)
      return 1;
   else if (l0<l1)
      return -1;

   return 0;
}


static void rank_vx(diagram_t *dm)
{
   int nv,nl,i,j,s;
   int (*v)[2];
   line_t *l;

   nv=(*dm).nv;
   v=(*dm).v;

   for (i=0;i<nv;i++)
      ivs[i]=i;

   for (i=0;i<(nv-1);i++)
   {
      for (j=(i+1);j<nv;j++)
      {
         if (cmp_vx(v,ivs[i],ivs[j])>0)
         {
            s=ivs[i];
            ivs[i]=ivs[j];
            ivs[j]=s;
         }
      }
   }

   for (i=0;i<nv;i++)
      rkv[ivs[i]]=i;

   nl=(*dm).nl;
   l=(*dm).l;

   for (i=0;i<nl;i++)
   {
      lnew[i].type=l[i].type;
      lold[i].type=l[i].type;
      lold[i].iv[0]=rkv[l[i].iv[0]];
      lold[i].iv[1]=rkv[l[i].iv[1]];
   }
}


static void set_ns(diagram_t *dm)
{
   int nv,ne,i,j;
   int (*v)[2];

   nv=(*dm).nv;
   v=(*dm).v;
   msi=0;
   mse=0;

   for (i=0;i<nv;i++)
   {
      ne=0;

      for (j=0;j<(int)(LTYPES);j++)
         ne+=eln[ivs[i]][j];

      if (ne==0)
      {
         if (msi==0)
         {
            ns[msi]=1;
            msi=1;
         }
         else if (cmp_vx(v,ivs[i-1],ivs[i])<0)
         {
            ns[msi]=1;
            msi+=1;
         }
         else
            ns[msi-1]+=1;
      }
      else
      {
         if (mse==0)
         {
            ns[msi]=1;
            mse=1;
         }
         else if (cmp_vx(v,ivs[i-1],ivs[i])<0)
         {
            ns[msi+mse]=1;
            mse+=1;
         }
         else
            ns[msi+mse-1]+=1;
      }
   }
}


static int cmp_line(line_t *l1,line_t *l2)
{
   if (((*l1).type==(*l2).type)&&
       ((*l1).iv[0]==(*l2).iv[0])&&((*l1).iv[1]==(*l2).iv[1]))
      return 1;

   if (((*l1).type==rvt[(int)((*l2).type)])&&
       ((*l1).iv[0]==(*l2).iv[1])&&((*l1).iv[1]==(*l2).iv[0]))
      return 1;

   return 0;
}


static int cmp_pdm(int nl)
{
   int i,j;

   for (i=0;i<nl;i++)
   {
      ils[i]=0;
      lnew[i].iv[0]=ips[lold[i].iv[0]];
      lnew[i].iv[1]=ips[lold[i].iv[1]];
   }

   for (i=0;i<nl;i++)
   {
      for (j=0;j<nl;j++)
      {
         if ((ils[j]==0)&&(cmp_line(lold+i,lnew+j)==1))
         {
            ils[j]=1;
            break;
         }
      }

      if (j==nl)
         return 0;
   }

   return 1;
}


static long factor1(diagram_t *dm)
{
   int nv,nl,i;
   long ip,np,fact;

   if (msi>0)
   {
      np=init_perm(msi,ns);

      if (np>1L)
      {
         nv=(*dm).nv;
         nl=(*dm).nl;

         for (i=0;i<nv;i++)
            ips[i]=i;

         ip=0L;
         fact=0L;

         while (ip<np)
         {
            ip=next_perm(ips);
            if (cmp_pdm(nl)==1)
               fact+=1L;
         }
      }
      else
         fact=1L;

      stop_perm();
   }
   else
      fact=1L;

   return fact;
}


static long factor2(diagram_t *dm)
{
   int nl,i,j,k;
   long fact,f;
   ltype_t lt;
   line_t *l;

   fact=1L;
   nl=(*dm).nl;
   l=(*dm).l;

   for (k=0;k<(int)(LTYPES);k++)
   {
      lt=(ltype_t)(k);

      if ((lt==SCALAR)||(lt==GLUON))
      {
         for (i=0;i<nl;i++)
         {
            if (l[i].type==lt)
            {
               ils[i]=1;

               for (j=0;j<i;j++)
               {
                  if ((ils[j]!=0)&&(cmp_line(l+i,l+j)==1))
                  {
                     ils[j]+=1;
                     ils[i]=0;
                     break;
                  }
               }
            }
            else
               ils[i]=0;
         }

         for (i=0;i<nl;i++)
         {
            if (ils[i]>1)
            {
               f=factorial(ils[i]);
               if (fact<=(LONG_MAX/f))
                  fact*=f;
               else
                  error(1,1,"factor2 [sfactor.c]",
                        "Symmetry factor is larger than LONG_MAX");
            }
         }
      }
   }

   return fact;
}


static long factor3(diagram_t *dm)
{
   int nl,i;
   long fact;
   line_t *l;

   nl=(*dm).nl;
   l=(*dm).l;
   fact=1L;

   for (i=0;i<nl;i++)
   {
      if (((l[i].type==SCALAR)||(l[i].type==GLUON))&&
          (l[i].iv[0]==l[i].iv[1]))
      {
         if (fact<=(LONG_MAX/2L))
            fact*=2L;
         else
            error(1,1,"factor3 [sfactor.c]",
                  "Symmetry factor is larger than LONG_MAX");
      }
   }

   return fact;
}


static long factor4(diagram_t *dm)
{
   int nv,nl,i,j;
   long ip,np,npi,npm;
   long f,fact;

   if (mse>0)
   {
      np=init_perm(msi+mse,ns);

      if (np>1L)
      {
         nl=(*dm).nl;
         npi=1L;

         for (i=0;i<msi;i++)
            npi*=factorial(ns[i]);

         ip=0L;
         fact=0L;
         npm=npi;

         while (ip<np)
         {
            ip=next_perm(ips);

            if (cmp_pdm(nl)==1)
            {
               fact+=1L;

               while (ip<npm)
                  ip=next_perm(ips);
            }

            if (ip==npm)
               npm+=npi;
         }
      }
      else
         fact=1L;

      stop_perm();
   }
   else
      fact=1L;

   nv=(*dm).nv;

   for (i=0;i<nv;i++)
   {
      for (j=0;j<(int)(LTYPES);j++)
      {
         if (eln[i][j]>1)
         {
            f=factorial(eln[i][j]);

            if (fact<=(LONG_MAX/f))
               fact*=f;
            else
               error(1,1,"factor4 [sfactor.c]",
                     "Symmetry factor is larger than LONG_MAX");
         }
      }
   }

   return fact;
}


long sfactor(theory_t th,diagram_t *dm,long *sf)
{
   int i;
   long fact;

   set_xln(th,dm);
   rank_vx(dm);
   set_ns(dm);

   sf[0]=factor1(dm);
   sf[1]=factor2(dm);
   sf[2]=factor3(dm);
   sf[3]=factor4(dm);

   fact=sf[0];

   for (i=1;i<4;i++)
   {
      if (fact<=(LONG_MAX/sf[i]))
         fact*=sf[i];
      else
         error(1,1,"sfactor [sfactor.c]",
               "Symmetry factor is larger than LONG_MAX");
   }

   return fact;
}
