
/*******************************************************************************
*
* File aatree.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generic Arne Andersson binary search trees.
*
* The externally accessible functions are
*
*   int init_aatree(void *context,
*                   int (*rank)(int oid1,int oid2,void *context),
*                   int (*is_equal)(int oid1,int oid2,void *context))
*     Initializes a new AA search tree with context "context", ranking
*     function "rank" and equality function "is_equal" (see the notes).
*     The program returns an integer search-tree identifier that is to
*     be used when invoking other functions in this module.
*
*   int insert_aatree(int oid,int tid)
*     Returns 0 if the object with index "oid" is already registered in
*     the search tree with identifier "tid". Otherwise the program adds
*     a node to the tree representing the object and returns 1.
*
*   int search_aatree(int oid,int tid,int *ifail)
*     Returns the index of an object registered in the search tree with
*     identifier "tid", which coincides with the object with index "oid".
*     The program returns oid if no such object is found. The argument
*     ifail is set to 0 if the object is found and to 1 otherwise.
*
*   void free_aatree(int tid)
*     Frees the nodes of the search tree with identifier "tid" through all
*     levels and resets the tree to its uninitialized state. No action is
*     performed if the tree is not initialized.
*
*   int check_aatree(int tid)
*     Checks whether the search tree with identifier "tid" is an AA tree
*     and returns the height of the tree. An error occurs if the tree is
*     not an AA tree.
*
* Notes:
*
* The programs in this module allow arrays of distinct objects to be built,
* which can be rapidly searched for a given object.
*
* First a context must be created, where the type of objects is defined and
* an object array is allocated. Knowing the context, it must be possible to
* access the objects in the array by giving the offset (also referred to as
* index or identifier) of the object from the base address of the array.
*
* Once the context is specified, a partial ordering of the objects must be
* defined through a function
*
*  int rank(int oid1,int oid2,void *context)
*
* that returns -1, 0 or 1 depending on whether the object with index oid1
* has rank less than, equal to or greater than the rank of the object with
* index oid2. Since the ordering is only a partial one, another function
*
*  int is_equal(int oid1,int oid2,void *context)
*
* must be provided, which returns 1 if the objects with indices oid1,oid2
* coincide and 0 otherwise. These programs must be such that is_equal()
* returns 0 if rank() returns -1 or 1.
*
* At this point, the search tree can be initialized by calling init_aatree()
* and may be built up recursively by adding objects one by one. A new object
* is added by temporarily appending it to the object array in the context at,
* say, offset "oid". It is then registered in the search tree by calling
*
*  insert_aatree(oid,tid),
*
* where "tid" is the identifier of the search tree returned by init_aatree().
* If insert_aatree() returns 0, the same object has previously been registered
* and the temporarily added object is removed from the array. Otherwise the
* object is kept and the next object is processed.
*
* An array built in this way contains only distinct objects. It can be searched
* for any given object by temporarily appending the object to the array at, say,
* offset "oid" and by calling
*
*  search_aatree(oid,tid,&ifail).
*
* If the object is found, ifail=0 and the program returns the index of that
* object in the array. Otherwise ifail=1 and "oid" is returned.
*
* The balancing algorithm for the search trees implemented in this module was
* introduced by Arne Andersson in the paper
*
*  "Balanced search trees made simple", in: "Workshop on algorithms and
*   data structures", p. 60, Springer Verlag (1993).
*
* Further details are given in the online tutorial
*
*  "Andersson trees", http://www.eternallyconfuzzled.com/tuts/
*                            datastructures/jsw_tut_andersson.aspx
*
* by Julienne Walker. The core functions are practically copied from there.
*
*******************************************************************************/

#define AATREE_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "dmlist.h"

typedef struct node
{
   int oid;
   int level;
   struct node *next[3];
} aanode_t;

typedef struct
{
   int init;
   void *context;
   int (*rank)(int oid1,int oid2,void *context);
   int (*is_equal)(int oid1,int oid2,void *context);
   aanode_t *root;
} aatree_t;

static int nts=0,nta=0;
static int flag,noid;
static aanode_t *nil;
static aatree_t *ats;


static void set_nil(void)
{
   nil=malloc(1*sizeof(*nil));
   error(nil==NULL,1,"set_nil [aatree.c]",
         "Unable to allocate level 0 node");

   (*nil).oid=0;
   (*nil).level=0;
   (*nil).next[0]=nil;
   (*nil).next[1]=nil;
   (*nil).next[2]=nil;
}


static aanode_t *new_node(int oid,int level)
{
   aanode_t *an;

   an=malloc(1*sizeof(*an));
   error(an==NULL,1,"new_node [aatree.c]","Unable to allocate new node");

   (*an).oid=oid;
   (*an).level=level;
   (*an).next[0]=nil;
   (*an).next[1]=nil;
   (*an).next[2]=nil;

   return an;
}


static void alloc_ats(void)
{
   int nt,i;
   aatree_t *at;

   if (nta==0)
      set_nil();

   nt=nta+32;
   at=malloc(nt*sizeof(*at));
   error(at==NULL,1,"alloc_ats [aatree.c]",
         "Unable to allocate new binary trees");

   for (i=0;i<nta;i++)
      at[i]=ats[i];

   for (i=nta;i<nt;i++)
   {
      at[i].init=0;
      at[i].context=NULL;
      at[i].rank=NULL;
      at[i].is_equal=NULL;
      at[i].root=nil;
   }

   if (nta>0)
      free(ats);

   ats=at;
   nta=nt;
}


static aanode_t *skew(aanode_t *root)
{
   aanode_t *left;

   if (root!=nil)
   {
      left=(*root).next[0];

      if ((*left).level==(*root).level)
      {
         (*root).next[0]=(*left).next[2];
         (*left).next[2]=root;
         return left;
      }
   }

   return root;
}


static aanode_t *split(aanode_t *root)
{
   aanode_t *right,*nright;

   if (root!=nil)
   {
      right=(*root).next[2];
      nright=(*right).next[2];

      if ((*nright).level==(*root).level)
      {
         (*root).next[2]=(*right).next[0];
         (*right).next[0]=root;
         (*right).level+=1;
         return right;
      }
   }

   return root;
}


static aanode_t *insert(aanode_t *root,int oid,int tid)
{
   int dir;
   aanode_t *an;
   aatree_t *at;

   if (root==nil)
   {
      root=new_node(oid,1);
      flag=1;
   }
   else
   {
      at=ats+tid;
      dir=(*at).rank(oid,(*root).oid,(*at).context)+1;

      if ((dir==0)||(dir==2))
      {
         (*root).next[dir]=insert((*root).next[dir],oid,tid);

         if (flag==1)
         {
            root=skew(root);
            root=split(root);
         }
      }
      else
      {
         an=root;

         while ((an!=nil)&&((*at).is_equal(oid,(*an).oid,(*at).context)!=1))
            an=(*an).next[1];

         if (an==nil)
         {
            an=new_node(oid,1);
            (*an).next[1]=(*root).next[1];
            (*root).next[1]=an;
            flag=2;
         }
      }
   }

   return root;
}


static void search(aanode_t *root,int oid,int tid)
{
   int dir;
   aanode_t *an;
   aatree_t *at;

   if (root!=nil)
   {
      at=ats+tid;
      dir=(*at).rank(oid,(*root).oid,(*at).context)+1;

      if ((dir==0)||(dir==2))
         search((*root).next[dir],oid,tid);
      else
      {
         an=root;

         while ((an!=nil)&&((*at).is_equal(oid,(*an).oid,(*at).context)!=1))
            an=(*an).next[1];

         if (an!=nil)
         {
            flag=0;
            noid=(*an).oid;
         }
      }
   }
}


int init_aatree(void *context,
                int (*rank)(int oid1,int oid2,void *context),
                int (*is_equal)(int oid1,int oid2,void *context))
{
   int nt;

   if (nts==nta)
      alloc_ats();

   nt=nts;
   ats[nt].init=1;
   ats[nt].context=context;
   ats[nt].rank=rank;
   ats[nt].is_equal=is_equal;
   ats[nt].root=nil;
   nts+=1;

   while ((nts<nta)&&(ats[nts].init==1))
      nts+=1;

   return nt;
}


int insert_aatree(int oid,int tid)
{
   flag=0;

   if ((tid>=0)&&(tid<nta)&&(ats[tid].init==1))
      ats[tid].root=insert(ats[tid].root,oid,tid);
   else
      error(1,1,"insert_aatree [aatree.c]",
            "Argument is out of range or the tree is not initialized");

   return (flag!=0);
}


int search_aatree(int oid,int tid,int *ifail)
{
   flag=1;
   noid=oid;

   if ((tid>=0)&&(tid<nta)&&(ats[tid].init==1))
      search(ats[tid].root,oid,tid);
   else
      error(1,1,"search_aatree [aatree.c]",
            "Argument is out of range or the tree is not initialized");

   (*ifail)=flag;
   return noid;
}


static void free_nodes(aanode_t *root)
{
   if (root==nil)
      return;

   free_nodes((*root).next[0]);
   free_nodes((*root).next[1]);
   free_nodes((*root).next[2]);
   free(root);
}


void free_aatree(int tid)
{
   if ((tid>=0)&&(tid<nta))
   {
      if (ats[tid].init==1)
      {
         free_nodes(ats[tid].root);

         ats[tid].init=0;
         ats[tid].context=NULL;
         ats[tid].rank=NULL;
         ats[tid].is_equal=NULL;
         ats[tid].root=nil;

         if (tid<nts)
            nts=tid;
      }
   }
   else
      error(1,1,"free_aatree [aatree.c]","Argument is out of range");
}


static int check_nodes(aanode_t *root)
{
   int hl,hr;
   int lev,llev,rlev,nlev;
   aanode_t *left,*right,*nright;

   if (root!=nil)
   {
      left=(*root).next[0];
      right=(*root).next[2];
      nright=(*right).next[2];

      lev=(*root).level;
      llev=(*left).level;
      rlev=(*right).level;
      nlev=(*nright).level;

      error((lev<=llev)||(lev<rlev)||(lev<=nlev)||
            ((lev>=2)&&((left==nil)||(right==nil))),1,
            "check_nodes [aatree.c]","Unexpected level pattern");

      hl=check_nodes(left);
      hr=check_nodes(right);

      if (hl>hr)
         return hl+1;
      else
         return hr+1;
   }
   else
      return 0;
}


int check_aatree(int tid)
{
   if ((tid>=0)&&(tid<nta))
   {
      if (ats[tid].init==1)
         return check_nodes(ats[tid].root);
   }
   else
      error(1,1,"check_aatree [aatree.c]","Argument is out of range");

   return 0;
}
