
/*******************************************************************************
*
* File dmlist.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Utility programs for diagram lists.
*
* The externally accessible functions are
*
*   dlist_t *alloc_dlist(int nmax,int flag)
*     Creates a new diagram list with initially space for nmax diagrams.
*     Vertex-rank and coefficient arrays are created depending on whether
*     the first and second bit of flag is set or not, respectively. A
*     search tree is initialized if and only if rank arrays are created.
*     The program returns the pointer to the new list.
*
*   void free_dlist(dlist_t *dl)
*     Frees the arrays in the diagram list dl, the search tree (if present)
*     and then the structure itself. The diagrams pointed to by the diagram
*     array, the arrays pointed to by the rank array and the coefficient
*     structures are not freed. The program does nothing if dl=NULL.
*
*   void deep_free_dlist(dlist_t *dl)
*     Frees the diagrams in the diagram list dl, the associated rank arrays,
*     the coefficient structures and then calls free_dlist(). The program
*     does nothing if dl=NULL.
*
*   void add_dgm(diagram_t *dm,dlist_t *dl)
*     Adds the diagram dm to a diagram list dl that was created w/o rank
*     arrays. An error occurs if the diagram list is not of this kind. If
*     the list was created with coefficient array, the coefficient of the
*     added diagram is set to 0. The diagram structure dm must have been
*     allocated using alloc_dgm() [diagram.c].
*
*   int insert_dgm(diagram_t *dm,int *rk,dlist_t *dl)
*     Adds the diagram dm to a diagram list dl that was created with rank
*     arrays. The diagram is not added if it is already contained in the
*     list. It must have been allocated using alloc_dgm() [diagram.c] and
*     is assumed to have been ordered by reorder_dgm() [diagrams/order.c],
*     the associated rank array being rk. The program returns 1 if the
*     diagram is added and 0 if it is already in the list.
*
*   int insert_copy(diagram_t *dm,int *rk,dlist_t *dl)
*     Same is insert_dgm() except that a copy of the diagram and the rank
*     array is made and added to the list if the diagram is not contained
*     in the list.
*
*   int search_dgm(diagram_t *dm,int *rk,dlist_t *dl)
*     Checks whether the diagram dm is contained in the diagram list dl
*     and returns the index of the diagram in the array (*dl).dma if it
*     is found. Otherwise the program returns (*dl).n. The diagram dm is
*     assumed to have been ordered by reorder_dgm() [diagrams/order.c],
*     the associated rank array being rk. An error occurs if the diagram
*     list was allocated w/o rank arrays.
*
* Notes:
*
* The structures of type dlist_t have the following elements:
*
*  n                     Current number of diagrams included in the
*                        list.
*
*  nmax                  Current maximal number of diagrams that can be
*                        included in the list. nmax and the arrays are
*                        adjusted dynamically when diagrams are added to
*                        the list.
*
*  tid                   Search tree identifier. Set to -1 if the diagram
*                        is created w/o rank arrays.
*
*  dma[i]                Pointer to the diagram number i (i=0,..,n-1).
*
*  rka[i]                Address of the vertex rank array of the i'th
*                        diagram (i=0,..,n-1) (rka=NULL if the structure
*                        was created without vertex rank arrays).
*
*  cna[i]                Address of the coefficient (of type cnum_t) of
*                        the i'th diagram (cna=NULL if the structure was
*                        created without coefficient arrays).
*
* Diagrams are described by structures of type diagram_t (see diagram.c).
* Unused elements of the diagram, vertex-rank and coefficient arrays are
* set to NULL.
*
* The type cnum_t is defined in utils.h and programs manipulating numbers
* of this type are collected in utils/cnum.c.
*
* Note that the addresses (*dl).dma, (*dl).rka and (*dl).cna may change
* when a diagram is added to the diagram list dl.
*
*******************************************************************************/

#define DMLIST_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "dmlist.h"

static long r0[1][2]={{0L,1L}};


static int rank(int oid1,int oid2,void *context)
{
   int **rka;
   diagram_t **dma;
   dlist_t *dl;

   dl=(dlist_t*)(context);
   rka=(*dl).rka;
   dma=(*dl).dma;

   return rank_dgm(dma[oid1],rka[oid1],dma[oid2],rka[oid2]);
}


static int is_equal(int oid1,int oid2,void *context)
{
   int **rka;
   diagram_t **dma;
   dlist_t *dl;

   dl=(dlist_t*)(context);
   rka=(*dl).rka;
   dma=(*dl).dma;

   return cmp_dgm(dma[oid1],rka[oid1],dma[oid2],rka[oid2]);
}


static diagram_t **new_dma(int nmax,int nold,diagram_t **dmold)
{
   int i;
   diagram_t **dma;

   dma=malloc(nmax*sizeof(*dma));
   error(dma==NULL,1,"new_dma [dmlist.c]",
         "Unable to allocate diagram array");

   for (i=0;i<nold;i++)
      dma[i]=dmold[i];

   for (i=nold;i<nmax;i++)
      dma[i]=NULL;

   if (nold>0)
      free(dmold);

   return dma;
}


static int **new_rka(int nmax,int nold,int **rkold)
{
   int i;
   int **rka;

   rka=malloc(nmax*sizeof(*rka));
   error(rka==NULL,1,"new_rka [dmlist.c]",
         "Unable to allocate vertex rank array");

   for (i=0;i<nold;i++)
      rka[i]=rkold[i];

   for (i=nold;i<nmax;i++)
      rka[i]=NULL;

   if (nold>0)
      free(rkold);

   return rka;
}


static cnum_t **new_cna(int nmax,int nold,cnum_t **cnold)
{
   int i;
   cnum_t **cna;

   cna=malloc(nmax*sizeof(*cna));
   error(cna==NULL,1,"new_cna [dmlist.c]",
         "Unable to allocate coefficient array");

   for (i=0;i<nold;i++)
      cna[i]=cnold[i];

   for (i=nold;i<nmax;i++)
      cna[i]=NULL;

   if (nold>0)
      free(cnold);

   return cna;
}


dlist_t *alloc_dlist(int nmax,int flag)
{
   dlist_t *dl;

   error(nmax<1,1,"alloc_dlist [dmlist.c]","Argument nmax is out of range");

   dl=malloc(1*sizeof(*dl));
   error(dl==NULL,1,"alloc_dlist [dmlist.c]","Unable to allocate diagram list");

   (*dl).n=0;
   (*dl).nmax=nmax;
   (*dl).dma=new_dma(nmax,0,NULL);

   if (flag&0x1)
   {
      (*dl).rka=new_rka(nmax,0,NULL);
      (*dl).tid=init_aatree((void*)(dl),rank,is_equal);
   }
   else
   {
      (*dl).rka=NULL;
      (*dl).tid=-1;
   }

   if (flag&0x2)
      (*dl).cna=new_cna(nmax,0,NULL);
   else
      (*dl).cna=NULL;

   return dl;
}


void free_dlist(dlist_t *dl)
{
   if (dl!=NULL)
   {
      free((*dl).dma);

      if ((*dl).rka!=NULL)
      {
         free((*dl).rka);
         free_aatree((*dl).tid);
      }

      if ((*dl).cna!=NULL)
         free((*dl).cna);

      free(dl);
   }
}


void deep_free_dlist(dlist_t *dl)
{
   int n,i;
   int **rka;
   cnum_t **cna;
   diagram_t **dma;

   if (dl!=NULL)
   {
      n=(*dl).n;
      dma=(*dl).dma;
      rka=(*dl).rka;
      cna=(*dl).cna;

      for (i=0;i<n;i++)
         free_dgm(dma[i]);

      if (rka!=NULL)
      {
         for (i=0;i<n;i++)
            free(rka[i]);
      }

      if (cna!=NULL)
      {
         for (i=0;i<n;i++)
            free_cnum(1,cna[i]);
      }

      free_dlist(dl);
   }
}


static void check_size(dlist_t *dl)
{
   int m,n,nmax;

   n=(*dl).n;
   nmax=(*dl).nmax;

   if (n==nmax)
   {
      if (nmax<=(INT_MAX/2))
         m=2*nmax;
      else
      {
         error(nmax==INT_MAX,1,"check_size [dmlist.c]",
               "Unable to extend diagram list");
         m=INT_MAX;
      }

      (*dl).nmax=m;
      (*dl).dma=new_dma(m,nmax,(*dl).dma);

      if ((*dl).rka!=NULL)
         (*dl).rka=new_rka(m,nmax,(*dl).rka);

      if ((*dl).cna!=NULL)
         (*dl).cna=new_cna(m,nmax,(*dl).cna);
   }
}


void add_dgm(diagram_t *dm,dlist_t *dl)
{
   int n;

   error((*dl).rka!=NULL,1,"add_dgm [dmlist.c]",
         "Cannot add diagram to a list with rank arrays");

   check_size(dl);
   n=(*dl).n;
   (*dl).n=n+1;
   (*dl).dma[n]=dm;

   if ((*dl).cna!=NULL)
   {
      if ((*dl).cna[n]==NULL)
         (*dl).cna[n]=alloc_cnum(1,0);
      else
         set_cnum(0,r0,(*dl).cna[n]);
   }
}


int insert_dgm(diagram_t *dm,int *rk,dlist_t *dl)
{
   int n;

   error((*dl).rka==NULL,1,"insert_dgm [dmlist.c]",
         "Cannot insert diagram to a list w/o rank arrays");

   check_size(dl);
   n=(*dl).n;
   (*dl).dma[n]=dm;
   (*dl).rka[n]=rk;

   if (insert_aatree(n,(*dl).tid)==1)
   {
      (*dl).n=n+1;

      if ((*dl).cna!=NULL)
      {
         if ((*dl).cna[n]==NULL)
            (*dl).cna[n]=alloc_cnum(1,0);
         else
            set_cnum(0,r0,(*dl).cna[n]);
      }

      return 1;
   }
   else
      return 0;
}


int insert_copy(diagram_t *dm,int *rk,dlist_t *dl)
{
   int n,i,*rkn;
   diagram_t *dmn;

   error((*dl).rka==NULL,1,"insert_copy [dmlist.c]",
         "Cannot insert diagram to a list w/o rank arrays");

   check_size(dl);
   n=(*dl).n;
   (*dl).dma[n]=dm;
   (*dl).rka[n]=rk;

   if (insert_aatree(n,(*dl).tid)==1)
   {
      dmn=alloc_dgm((*dm).nv,(*dm).nl);
      rkn=alloc_rank((*dm).nv);
      copy_dgm(dm,dmn);

      for (i=0;i<(*dm).nv;i++)
         rkn[i]=rk[i];

      (*dl).n=n+1;
      (*dl).dma[n]=dmn;
      (*dl).rka[n]=rkn;

      if ((*dl).cna!=NULL)
      {
         if ((*dl).cna[n]==NULL)
            (*dl).cna[n]=alloc_cnum(1,0);
         else
            set_cnum(0,r0,(*dl).cna[n]);
      }

      return 1;
   }
   else
      return 0;
}


int search_dgm(diagram_t *dm,int *rk,dlist_t *dl)
{
   int n,ifail;

   error((*dl).rka==NULL,1,"search_dgm [dmlist.c]",
         "Cannot search a diagram in a list w/o rank arrays");

   check_size(dl);
   n=(*dl).n;
   (*dl).dma[n]=dm;
   (*dl).rka[n]=rk;

   return search_aatree(n,(*dl).tid,&ifail);
}
