
/*******************************************************************************
*
* File ptools.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Basic plot functions.
*
* The externally accessible functions are
*
*   void start_plot(FILE *plot,int nc)
*     Writes the required latex header commands to the stream "plot"
*     and moves the pointer to the origin of the plot area. The latter
*     is logically divided into "nc" columns in which the diagrams
*     will be arranged. This program must be called before any plot
*     data are written. All plot commands go to the specified stream
*     until the plot ends.
*
*   void end_plot(void)
*     Writes the required closing latex commands and ends the plot.
*     This program must be called before a new plot is started.
*
*   void set_point(double *pos)
*     Sets the pointer position to the point with Cartesian coordinates
*     (pos[0],pos[1]) with respect to the origin at the lower-left corner
*     of the plot area [which has coordinates (0.0,0.0)].
*
*   void get_point(double *pos)
*     Assigns the current Cartesian coordinates of the pointer position
*     to pos[0],pos[1].
*
*   void move_point(double phi,double *vec)
*     Moves the pointer from the current position by the displacement
*     vector R(phi)*vec, where R(phi) rotates the vector by the angle
*     phi. The angle has to be given in degrees and the rotation goes
*     in the counterclockwise direction.
*
*   void draw_line(ltype_t lt,double phi,double *vec)
*     Draws a line of type "lt" from the current pointer position pos
*     to pos+R(phi)*vec [see move_point()]. The pointer is not moved.
*
*   void draw_labeled_line(ltype_t lt,double phi,double *vec,int lbl)
*     Draws a line of type "lt" from the current pointer position pos
*     to pos+R(phi)*vec [see move_point()]. Then adds a label "lbl" at
*     the end of the line. The pointer is not moved.
*
*   void draw_rooted_line(ltype_t lt,double phi,double *vec)
*     Draws a line of type "lt" from the current pointer position pos
*     to pos+R(phi)*vec [see move_point()]. Then adds a little square
*     at the end of the line. The pointer is not moved.
*
*   void draw_vertex(vtype_t vt)
*     Draws a vertex of type "vt" at the current pointer position (see
*     the notes).
*
*   void draw_coefficient(cnum_t *cn)
*     Draws the value of the coefficient cn at the current pointer
*     position and moves the pointer to the end of the text.
*
*   void start_bbox(void)
*     Drawing commands that follow a call to this program are not written
*     to the plot file. The commands are however executed internally and
*     are used to determine the bounding box of the diagram that would be
*     drawn. This mode of operation lasts until end_bbox() is called.
*
*   void end_bbox(double *pos1,double *pos2)
*     Exits the mode of operation started by start_bbox() and resets the
*     pointer position to where it was when start_bbox() was called. The
*     lower-left and upper-right positions of the bounding box containing
*     the diagram elements that would have been drawn are assigned to
*     pos1 and pos2.
*
*   void start_diagram(double *pos1,double *pos2)
*     Moves the pointer position to a point in the plot area, where the
*     a diagram with bounding box given by pos1 and pos2 can be drawn.
*     The program assumes that the current pointer position is such that
*     the diagram would fit into the box with lower-left corner pos1 and
*     upper-right corner pos2 if it were drawn starting from this point.
*
*   void end_diagram(void)
*     Exits the mode of operation started by start_diagram() and moves
*     the pointer position to the origin.
*
* Notes:
*
* These programs write LaTeX plot commands to a file that can be processed
* using latex followed by dvips. The commands require the axodraw.sty LaTeX
* style file (see doc/axodraw.pdf).
*
* The graphical symbols for the vertices are
*
*  SOURCE                Large open circle.
*
*  EXTERNAL              Open square.
*
*  ACTION                Filled circle.
*
*  MEASURE               Filled square.
*
*  GFIX                  Open circle.
*
*  MASS                  Plus sign rotated by 45 degrees.
*
* For the gluon a wavy line is used rather than the standard curled line.
*
*******************************************************************************/

#define PTOOLS_C

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "utils.h"
#include "plot.h"

static int ncol,icol,ibb=0,idm=0;
static double base,width,height;
static double ptpos[2],ptold[2],bb1[2],bb2[2];
static FILE *out=NULL;


void start_plot(FILE *plot,int nc)
{
   if (out==NULL)
      out=plot;
   else
      error(1,1,"start_plot [ptools.c]",
            "Attempt to start a new plot before ending the current one");

   error(nc<1,1,"start_plot [ptools.c]",
         "Number of columns is out of range");

   ncol=nc;
   icol=0;
   base=YSIZE;
   width=XSIZE/(double)(ncol);
   height=0.0;

   ptpos[0]=0.0;
   ptpos[1]=0.0;

   fprintf(out,"\\documentclass{article}\n");
   fprintf(out,"\\usepackage{geometry}\n");
   fprintf(out,"\\usepackage{axodraw}\n");
   fprintf(out,"\\usepackage{nopageno}\n");
   fprintf(out,"\\geometry{a4paper,left=20mm,bottom=20mm,\n");
   fprintf(out,"          textwidth=170mm,textheight=257mm}\n");
   fprintf(out,"\\begin{document}\n");
   fprintf(out,"\\setlength{\\unitlength}{1bp}\n");

   fprintf(out,"\\begin{center}\n");
   fprintf(out,"\\begin{picture}(%.2f,%.2f)(0.0,0.0)\n",XSIZE,YSIZE);
   fprintf(out,"\\SetPFont{Times-Roman}{%d}\n",(int)(TXTSIZE+0.5));
}


void end_plot(void)
{
   fprintf(out,"\\end{picture}\n");
   fprintf(out,"\\end{center}\n");
   fprintf(out,"\\end{document}\n");
   out=NULL;
}


static void new_picture(void)
{
   fprintf(out,"\\end{picture}\n\n");
   fprintf(out,"\\end{center}\n\n");
   fprintf(out,"\\begin{center}\n");
   fprintf(out,"\\begin{picture}(%.2f,%.2f)(0.0,0.0)\n",XSIZE,YSIZE);
   fprintf(out,"\\SetPFont{Times-Roman}{%d}\n",(int)(TXTSIZE+0.5));

   icol=0;
   base=YSIZE;
   height=0.0;
}


void set_point(double *pos)
{
   ptpos[0]=pos[0];
   ptpos[1]=pos[1];
}


void get_point(double *pos)
{
   pos[0]=ptpos[0];
   pos[1]=ptpos[1];
}


void move_point(double phi,double *vec)
{
   double c,s;

   phi*=(3.141592654/180.0);
   c=cos(phi);
   s=sin(phi);

   ptpos[0]+=(c*vec[0]-s*vec[1]);
   ptpos[1]+=(s*vec[0]+c*vec[1]);
}


static void update_bb(double *pos)
{
   if (pos[0]<bb1[0])
      bb1[0]=pos[0];
   else if (pos[0]>bb2[0])
      bb2[0]=pos[0];

   if (pos[1]<bb1[1])
      bb1[1]=pos[1];
   else if (pos[1]>bb2[1])
      bb2[1]=pos[1];
}


void draw_line(ltype_t lt,double phi,double *vec)
{
   double c,s,nw,end[2];

   if (lt!=LTYPES)
   {
      phi*=(3.141592654/180.0);
      c=cos(phi);
      s=sin(phi);
      nw=4.0*sqrt(vec[0]*vec[0]+vec[1]*vec[1])/20.0;

      end[0]=ptpos[0]+(c*vec[0]-s*vec[1]);
      end[1]=ptpos[1]+(s*vec[0]+c*vec[1]);

      if (!ibb)
      {
         if (lt==SCALAR)
            fprintf(out,"\\Line(%.2f,%.2f)(%.2f,%.2f)\n",
                    ptpos[0],ptpos[1],end[0],end[1]);
         else if (lt==GLUON)
            fprintf(out,"\\Photon(%.2f,%.2f)(%.2f,%.2f){1.5}{%.2f}\n",
                    ptpos[0],ptpos[1],end[0],end[1],nw);
         else if (lt==GHOST)
            fprintf(out,"\\DashArrowLine(%.2f,%.2f)(%.2f,%.2f){1}\n",
                    ptpos[0],ptpos[1],end[0],end[1]);
         else if (lt==ANTIGHOST)
            fprintf(out,"\\DashArrowLine(%.2f,%.2f)(%.2f,%.2f){1}\n",
                    end[0],end[1],ptpos[0],ptpos[1]);
         else if (lt==QUARK)
            fprintf(out,"\\ArrowLine(%.2f,%.2f)(%.2f,%.2f)\n",
                    ptpos[0],ptpos[1],end[0],end[1]);
         else if (lt==ANTIQUARK)
            fprintf(out,"\\ArrowLine(%.2f,%.2f)(%.2f,%.2f)\n",
                    end[0],end[1],ptpos[0],ptpos[1]);
      }
      else
      {
         update_bb(ptpos);
         update_bb(end);
      }
   }
}


void draw_labeled_line(ltype_t lt,double phi,double *vec,int lbl)
{
   double c,s,tpos[2];
   double r,n,end[2];

   if (lt!=LTYPES)
   {
      draw_line(lt,phi,vec);

      phi*=(3.141592654/180.0);
      c=cos(phi);
      s=sin(phi);
      n=sqrt(vec[0]*vec[0]+vec[1]*vec[1]);
      r=2.5*SCALE_FACTOR;

      end[0]=ptpos[0]+(c*vec[0]-s*vec[1]);
      end[1]=ptpos[1]+(s*vec[0]+c*vec[1]);

      tpos[0]=end[0]+(r+TXTSEP+    TXTUP)*(c*vec[0]-s*vec[1])/n;
      tpos[1]=end[1]+(r+TXTSEP+2.0*TXTUP)*(s*vec[0]+c*vec[1])/n+0.7*TXTUP;

      if (!ibb)
      {
         fprintf(out,"\\GCirc(%.2f,%.2f){%.1f}{1}\n",
                 end[0],end[1],r);
         if (lbl>0)
         {
            fprintf(out,"\\PText(%.2f,%.2f)(0)[]{%d}\n",
                    tpos[0],tpos[1],lbl);
         }
      }
      else
      {
         end[0]-=r;
         end[1]-=r;
         update_bb(end);
         end[0]+=2.0*r;
         end[1]+=2.0*r;
         update_bb(end);

         if (lbl>0)
         {
            update_bb(tpos);
            tpos[0]+=TXTSIZE;
            tpos[1]+=TXTSIZE;
            update_bb(tpos);
         }
      }
   }
}


void draw_rooted_line(ltype_t lt,double phi,double *vec)
{
   double c,s;
   double r,end[2];

   if (lt!=LTYPES)
   {
      draw_line(lt,phi,vec);

      phi*=(3.141592654/180.0);
      c=cos(phi);
      s=sin(phi);

      r=2.5*SCALE_FACTOR;
      end[0]=ptpos[0]+(c*vec[0]-s*vec[1]);
      end[1]=ptpos[1]+(s*vec[0]+c*vec[1]);

      if (!ibb)
      {
         fprintf(out,"\\GBoxc(%.2f,%.2f)(%.1f,%.1f){1}\n",
                 end[0],end[1],r,r);
      }
      else
      {
         update_bb(ptpos);
         end[0]-=0.5*r;
         end[1]-=0.5*r;
         update_bb(end);
         end[0]+=r;
         end[1]+=r;
         update_bb(end);
      }
   }
}


void draw_vertex(vtype_t vt)
{
   double r,end[2];

   if (vt==SOURCE)
   {
      r=1.5*SCALE_FACTOR;

      if (!ibb)
         fprintf(out,"\\GCirc(%.2f,%.2f){%.1f}{0}\n",
                 ptpos[0],ptpos[1],r);
   }
   else if (vt==EXTERNAL)
   {
      r=2.5*SCALE_FACTOR;

      if (!ibb)
         fprintf(out,"\\GBoxc(%.2f,%.2f)(%.1f,%.1f){1}\n",
                 ptpos[0],ptpos[1],r,r);

      r*=0.5;
   }
   else if (vt==ACTION)
   {

      r=1.5*SCALE_FACTOR;

      if (!ibb)
         fprintf(out,"\\GCirc(%.2f,%.2f){%.1f}{0}\n",
                 ptpos[0],ptpos[1],r);
   }
   else if (vt==MASS)
   {
      r=2.0*SCALE_FACTOR;

      if (!ibb)
      {
         fprintf(out,"\\Line(%.2f,%.2f)(%.2f,%.2f)",
                 ptpos[0]-r,ptpos[1]-r,ptpos[0]+r,ptpos[1]+r);
         fprintf(out,"\\Line(%.2f,%.2f)(%.2f,%.2f)\n",
                 ptpos[0]-r,ptpos[1]+r,ptpos[0]+r,ptpos[1]-r);
      }

      r/=sqrt(2.0);
   }
   else if (vt==GFIX)
   {
      r=2.5*SCALE_FACTOR;

      if (!ibb)
         fprintf(out,"\\GCirc(%.2f,%.2f){%.1f}{0.8}\n",
                 ptpos[0],ptpos[1],r);
   }
   else if (vt==MEASURE)
   {
      r=4.0*SCALE_FACTOR;

      if (!ibb)
         fprintf(out,"\\GBoxc(%.2f,%.2f)(%.1f,%.1f){0.8}\n",
                 ptpos[0],ptpos[1],r,r);

      r*=0.5;
   }
   else
      r=0.0;

   if ((ibb)&&(vt!=VTYPES))
   {
      end[0]=ptpos[0]+r;
      end[1]=ptpos[1]+r;
      update_bb(end);
      end[0]=ptpos[0]-r;
      end[1]=ptpos[1]-r;
      update_bb(end);
   }
}


void draw_coefficient(cnum_t *cn)
{
   int nw;
   double r,end[2];

   if (!ibb)
   {
      fprintf(out,"\\PText(%.2f,%.2f)(0)[l]{",
              ptpos[0],ptpos[1]);
      nw=print_cnum(out,"Nf",cn);
      fprintf(out,"}\n");
      ptpos[0]+=nw*0.65*TXTSIZE;
   }
   else
   {
      r=TXTSIZE;
      end[0]=ptpos[0]-r;
      end[1]=ptpos[1]-r;
      update_bb(end);
      end[0]=ptpos[0]+r;
      end[1]=ptpos[1]+r;
      update_bb(end);

      nw=print_cnum(out,"Nf",cn);
      fseek(out,(long)(-nw),SEEK_CUR);
      ptpos[0]+=(double)(nw)*0.65*TXTSIZE;

      end[0]=ptpos[0]-r;
      end[1]=ptpos[1]-r;
      update_bb(end);
      end[0]=ptpos[0]+r;
      end[1]=ptpos[1]+r;
      update_bb(end);
   }
}


void start_bbox(void)
{
   error(ibb,1,"start_bbox [ptools.c]",
         "Bounding box mode is already switched on");

   ibb=1;
   ptold[0]=ptpos[0];
   ptold[1]=ptpos[1];

   bb1[0]=ptpos[0];
   bb1[1]=ptpos[1];

   bb2[0]=ptpos[0];
   bb2[1]=ptpos[1];
}


void end_bbox(double *pos1,double *pos2)
{
   error(!ibb,1,"end_bbox [ptools.c]",
         "Bounding box mode is already switched off");

   pos1[0]=bb1[0]-1.5;
   pos1[1]=bb1[1]-1.5;

   pos2[0]=bb2[0]+1.5;
   pos2[1]=bb2[1]+1.5;

   ptpos[0]=ptold[0];
   ptpos[1]=ptold[1];
   ibb=0;
}


void start_diagram(double *pos1,double *pos2)
{
   double d[2],s[2];

   error(idm,1,"start_diagram [ptools.c]",
         "Diagram mode is already switched on");

   idm=1;
   d[0]=pos2[0]-pos1[0];
   d[1]=pos2[1]-pos1[1];

   error((d[0]>XSIZE)||(d[1]>YSIZE),1,"start_diagram [ptools.c]",
         "Diagram does not fit into plot area");

   if (d[1]>base)
      new_picture();

   if (d[0]>((double)(ncol-icol)*width))
   {
      if ((d[1]+height+VSPACE)>base)
         new_picture();
      else
      {
         base-=(height+VSPACE);
         height=0.0;
         icol=0;
      }
   }

   s[0]=(double)(icol)*width-pos1[0];
   s[1]=base-pos2[1];

   ptpos[0]+=s[0];
   ptpos[1]+=s[1];

   if (d[1]>height)
      height=d[1];

   icol+=(int)(ceil((d[0]+HSPACE)/width));
}


void end_diagram(void)
{
   error(!idm,1,"end_diagram [ptools.c]",
         "Diagram mode is already switched off");

   if (!ibb)
      fprintf(out,"\n");

   ptpos[0]=0.0;
   ptpos[1]=0.0;
   idm=0;
}
